@@ -267,6 +267,8 @@
      (url . ,expand-url)
      (hyperlink . ,expand-hyperlink)
      (rawcode . code)
+     (pre . pre)
+     (bibitem . ,(lambda (x env) '()))  ;; TODO: bibtex
      (code . ,expand-code)
      (codeblock . ,expand-codeblock)
      (ccode
@@ -1039,7 +1041,8 @@ h4 { color: #222288; border-top: 1px solid #4588ba; }
 (define (extract-module-docs-from-files mod srcs includes stubs strict? exports . o)
   (let ((dir (or (and (pair? o) (car o)) (module-dir mod)))
         (defs (map (lambda (x)
-                     (let ((val (and mod (module-ref mod x))))
+                     (let ((val (and mod (protect (exn (else #f))
+                                           (module-ref mod x)))))
                        `(,x ,val ,(object-source val))))
                    exports)))
     (define (resolve-file file)
