@@ -9,25 +9,10 @@
   (call-with-input-string " "
     (lambda (in) (read-char in) (read-char in))))
 
-;; Copy whole characters from the given cursor positions.
-;; Return the src cursor position of the next unwritten char,
-;; which may be before `to' if the char would overflow.
-;; Now provided as a primitive from (chibi ast).
-;; (define (string-cursor-copy! dst start src from to)
-;;   (let lp ((i from)
-;;            (j (string-cursor->index dst start)))
-;;     (let ((i2 (string-cursor-next src i)))
-;;       (cond ((> i2 to) i)
-;;             (else
-;;              (string-set! dst j (string-cursor-ref src i))
-;;              (lp i2 (+ j 1)))))))
-
 (define (utf8->string vec . o)
-  (if (pair? o)
-      (let ((start (car o))
-            (end (if (pair? (cdr o)) (cadr o) (bytevector-length vec))))
-        (utf8->string (subbytes vec start end)))
-      (string-copy (utf8->string! vec))))
+  (let ((start (if (pair? o) (car o) 0))
+        (end (if (and (pair? o) (pair? (cdr o))) (cadr o) (bytevector-length vec))))
+    (string-copy (utf8->string! vec start end))))
 
 (define (string->utf8 str . o)
   (if (pair? o)
@@ -75,9 +60,12 @@
 ;;> a string not including the newline.  Reads at most \var{n}
 ;;> characters, defaulting to 8192.
 
-(cond-expand
- ((not string-streams)
-  (define (%read-line n in)
+(define (%read-line n in)
+  (cond
+   ((stream-port? in) ;;(port-fileno in)
+    (port-line-set! in (+ 1 (port-line in)))
+    (%%read-line n in))
+   (else
     (let ((out (open-output-string)))
       (let lp ((i 0))
         (let ((ch (peek-char in)))
@@ -103,13 +91,10 @@
   (let ((in (if (pair? o) (car o) (current-input-port)))
         (n (if (and (pair? o) (pair? (cdr o))) (car (cdr o)) 8192)))
     (let ((res (%read-line n in)))
-      (cond-expand
-       (string-streams
-        (port-line-set! in (+ 1 (port-line in)))))
       (if (not res)
           eof
           (let ((len (string-length res)))
-            (cond
+            (cond  ;; strip crlf
              ((and (> len 0) (eqv? #\newline (string-ref res (- len 1))))
               (if (and (> len 1) (eqv? #\return (string-ref res (- len 2))))
                   (substring res 0 (- len 2))
@@ -128,9 +113,11 @@
 ;;> than \var{n} characters if the end of file is reached,
 ;;> or the eof-object if no characters are available.
 
-(cond-expand
- ((not string-streams)
-  (define (%read-string n in)
+(define (%read-string n in)
+  (cond
+   ;;((port-fileno in)
+   ;; (%%read-string n in))
+   (else
     (let ((out (open-output-string)))
       (let lp ((i 0))
         (cond ((or (= i n) (eof-object? (peek-char in)))
@@ -159,9 +146,11 @@
 ;;> An error is signalled if the length of \var{str} is smaller
 ;;> than \var{n}.
 
-(cond-expand
- ((not string-streams)
-  (define (%read-string! str n in)
+(define (%read-string! str n in)
+  (cond
+   ;;((port-fileno in)
+   ;; (%%read-string! str n in))
+   (else
     (let lp ((i 0))
       (cond ((or (= i n) (eof-object? (peek-char in))) i)
             (else (string-set! str i (read-char in)) (lp (+ i 1))))))))
