@@ -24,15 +24,25 @@
 ;; most of these are plain text for easier viewing in the browser
 (define (mime-type-from-extension ext)
   (assq-ref
-   '((htm . "text/html; charset=utf-8")
+   '((c . "text/plain; charset=utf-8")
+     (css . "text/css; charset=utf-8")
+     (gif . "image/gif")
+     (h . "text/plain; charset=utf-8")
+     (htm . "text/html; charset=utf-8")
      (html . "text/html; charset=utf-8")
+     (jpeg . "image/jpeg")
+     (jpg . "image/jpeg")
+     (js . "application/javascript; charset=utf-8")
+     (json . "application/json; charset=utf-8")
+     (md . "text/plain; charset=utf-8")
+     (mp3 . "audio/mpeg")
+     (org . "text/plain; charset=utf-8")
+     (pdf . "application/pdf")
+     (png . "image/png")
      (scm . "text/plain; charset=utf-8")
      (sld . "text/plain; charset=utf-8")
-     (c . "text/plain; charset=utf-8")
-     (h . "text/plain; charset=utf-8")
-     (txt . "text/plain; charset=utf-8")
-     (org . "text/plain; charset=utf-8")
-     (md . "text/plain; charset=utf-8"))
+     (svg . "image/svg+xml")
+     (txt . "text/plain; charset=utf-8"))
    (and (string? ext) (string->symbol ext))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
