@@ -24,9 +24,9 @@
    #u8(0 0 0 0 0 0 0 0 0 0 0 0)))
 
 (define (file-owner-or-nobody uid)
-  (or (user-name (user-information uid)) "nobody"))
+  (or (cond ((user-information uid) => user-name) (else #f)) "nobody"))
 (define (file-group-or-nobody gid)
-  (or (group-name (group-information gid)) "nobody"))
+  (or (cond ((group-information gid) => group-name) (else #f)) "nobody"))
 
 (define (make-tar file mode uid gid size mod-time type . o)
   (let* ((link (if (pair? o) (car o) ""))
@@ -280,17 +280,18 @@
         res))))
 
 (define (main args)
-  (cond
-    ((equal? "t" (car args))
-     (for-each (lambda (f) (write-string f) (newline)) (tar-files (cadr args))))
-    ((equal? "x" (car args))
-     (if (tar-safe? (cadr args))
-       (tar-extract (cadr args))
-       (error "tar file not a single relative directory" (cadr args))))
-    ((equal? "c" (car args))
-     (tar-create (cadr args) (cddr args)))
-    ((equal? "f" (car args))
-     (write-string
+  (let ((args (cdr args)))
+    (cond
+     ((equal? "t" (car args))
+      (for-each (lambda (f) (write-string f) (newline)) (tar-files (cadr args))))
+     ((equal? "x" (car args))
+      (if (tar-safe? (cadr args))
+          (tar-extract (cadr args))
+          (error "tar file not a single relative directory" (cadr args))))
+     ((equal? "c" (car args))
+      (tar-create (cadr args) (cddr args)))
+     ((equal? "f" (car args))
+      (write-string
        (utf8->string (tar-extract-file (cadr args) (car (cddr args))))))
-    (else
-     (error "unknown tar command" (car args)))))
+     (else
+      (error "unknown tar command" (car args))))))
