@@ -17,8 +17,8 @@
 (define fl- -)
 (define fl* *)
 (define fl/ /)
-(define flmax max)
-(define flmin min)
+(define (flmax . args) (if (null? args) -inf.0 (apply max args)))
+(define (flmin . args) (if (null? args) +inf.0 (apply min args)))
 (define (flabsdiff x y) (abs (- x y)))
 (define flnumerator numerator)
 (define fldenominator denominator)
