@@ -132,15 +132,12 @@
                (vector? sizes)
                (= (array-dimension array) (vector-length sizes))
                (vector-every exact-integer? sizes)
-               (vector-every >= sizes (interval-lower-bounds->vector
-                                       (array-domain array)))
-               (vector-every < sizes (interval-upper-bounds->vector
-                                      (array-domain array)))))
+               (vector-every <= sizes (interval-ub (array-domain array)))))
   (let ((domain (make-interval
                  (vector-map
                   (lambda (lo hi s) (exact (ceiling (/ (- hi lo) s))))
-                  (interval-lower-bounds->vector (array-domain array))
-                  (interval-upper-bounds->vector (array-domain array))
+                  (interval-lb (array-domain array))
+                  (interval-ub (array-domain array))
                   sizes))))
     (make-array
      domain
@@ -150,13 +147,13 @@
         (make-interval
          (vector-map
           (lambda (i lo s) (+ lo (* i s)))
-          multi-index
-          (interval-lower-bound (array-domain array))
+          (list->vector multi-index)
+          (interval-lb (array-domain array))
           sizes)
          (vector-map
           (lambda (i lo hi s)
             (min hi (+ lo (* (+ i 1) s))))
-          multi-index
+          (list->vector multi-index)
           (interval-lb (array-domain array))
           (interval-ub (array-domain array))
           sizes)))))))
@@ -291,6 +288,12 @@
       (and (equal? (array-dimension (car ls)) (array-dimension (cadr ls)))
            (same-dimensions? (cdr ls)))))
 
+(define (same-domains? ls)
+  (or (null? ls)
+      (null? (cdr ls))
+      (and (interval= (array-domain (car ls)) (array-domain (cadr ls)))
+           (same-domains? (cdr ls)))))
+
 (define (array-map f array . arrays)
   (make-array (array-domain array)
               (let* ((ls (cons array arrays))
@@ -300,13 +303,32 @@
                   (apply f (map (lambda (g) (apply g multi-index)) getters))))))
 
 (define (array-for-each f array . arrays)
-  (interval-for-each
-   (let* ((ls (cons array arrays))
-          (getters (map array-getter ls)))
-     (assert (same-dimensions? ls))
-     (lambda multi-index
-       (apply f (map (lambda (g) (apply g multi-index)) getters))))
-   (array-domain array)))
+  (if (null? arrays)
+      (interval-for-each
+       (let ((g (array-getter array)))
+         (case (array-dimension array)
+           ((1)
+            (lambda (i) (f (g i))))
+           ((2)
+            (lambda (i j) (f (g i j))))
+           (else
+            (lambda multi-index
+              (f (apply g multi-index))))))
+       (array-domain array))
+      (interval-for-each
+       (let* ((lower (interval-lower-bounds->list (array-domain array)))
+              (ls (cons array arrays))
+              (getters
+               (cons (array-getter (car ls))
+                     (map (lambda (ar)
+                            (let ((getter (array-getter ar)))
+                              (lambda multi-index
+                                (apply getter multi-index))))
+                          (cdr ls)))))
+         (assert (same-domains? ls))
+         (lambda multi-index
+           (apply f (map (lambda (g) (apply g multi-index)) getters))))
+       (array-domain array))))
 
 (define (array-fold kons knil array)
   (interval-fold (lambda (acc . multi-index)
