@@ -3,6 +3,17 @@
 ;; This code was written by Alex Shinn in 2009 and placed in the
 ;; Public Domain.  All warranties are disclaimed.
 
+(define escaped-chars
+  '((#\alarm . "alarm")
+    (#\backspace . "backspace")
+    (#\delete . "delete")
+    (#\escape . "escape")
+    (#\newline . "newline")
+    (#\null . "null")
+    (#\return . "return")
+    (#\space . "space")
+    (#\tab . "tab")))
+
 (define (raise-typed-error type)
   (lambda (msg . args) (raise (make-exception type msg args #f #f))))
 (define read-error (raise-typed-error 'read))
@@ -66,60 +77,65 @@
                       (set! count (+ count 1))))
                (cont x index)))))
     (let wr ((x x))
-      (check-shared
-       x
-       ""
-       (lambda (x shared?)
-         (cond
-          ((pair? x)
-           (display "(" out)
-           (wr (car x))
-           (let lp ((ls (cdr x)))
-             (check-shared
-              ls
-              " . "
-              (lambda (ls shared?)
-                (cond ((null? ls))
-                      ((pair? ls)
-                       (cond
-                        (shared?
-                         (display "(" out)
-                         (wr (car ls))
-                         (check-shared
-                          (cdr ls)
-                          " . "
-                          (lambda (ls shared?) (lp ls)))
-                         (display ")" out))
-                        (else
-                         (display " " out)
-                         (wr (car ls))
-                         (lp (cdr ls)))))
-                      (else
-                       (display " . " out)
-                       (wr ls))))))
-           (display ")" out))
-          ((vector? x)
-           (display "#(" out)
-           (let ((len (vector-length x)))
-             (cond ((> len 0)
-                    (wr (vector-ref x 0))
-                    (do ((i 1 (+ i 1)))
-                        ((= i len))
-                      (display " " out)
-                      (wr (vector-ref x i))))))
-           (display ")" out))
-          ((let ((type (type-of x)))
-             (and (type? type) (type-printer type)))
-           => (lambda (printer) (printer x wr out)))
-          ((null? x) (display "()" out))
-          ((char? x) (display "#\\" out) (write-char x out))
-          ((symbol? x) (write x out))
-          ((number? x) (display (number->string x) out))
-          ((eq? x #t) (display "#t" out))
-          ((eq? x #f) (display "#f" out))
-          (else
-           ;; (display "#<unknown>" out)
-           (write x out))))))))
+      (define (wr-one x shared?)
+        (cond
+         ((pair? x)
+          (display "(" out)
+          (wr (car x))
+          (let lp ((ls (cdr x)))
+            (check-shared
+             ls
+             " . "
+             (lambda (ls shared?)
+               (cond ((null? ls))
+                     ((pair? ls)
+                      (cond
+                       (shared?
+                        (display "(" out)
+                        (wr (car ls))
+                        (check-shared
+                         (cdr ls)
+                         " . "
+                         (lambda (ls shared?) (lp ls)))
+                        (display ")" out))
+                       (else
+                        (display " " out)
+                        (wr (car ls))
+                        (lp (cdr ls)))))
+                     (shared?  ;; shared dotted tail
+                      (wr-one ls #f))
+                     (else
+                      (display " . " out)
+                      (wr ls))))))
+          (display ")" out))
+         ((vector? x)
+          (display "#(" out)
+          (let ((len (vector-length x)))
+            (cond ((> len 0)
+                   (wr (vector-ref x 0))
+                   (do ((i 1 (+ i 1)))
+                       ((= i len))
+                     (display " " out)
+                     (wr (vector-ref x i))))))
+          (display ")" out))
+         ((let ((type (type-of x)))
+            (and (type? type) (type-printer type)))
+          => (lambda (printer) (printer x wr out)))
+         ((null? x) (display "()" out))
+         ((char? x)
+          (display "#\\" out)
+          (let ((pair (assv x escaped-chars)))
+            (if pair
+              (display (cdr pair) out)
+              (write-char x out))))
+         ((symbol? x) (write x out))
+         ((number? x) (display (number->string x) out))
+         ((eq? x #t) (display "#t" out))
+         ((eq? x #f) (display "#f" out))
+         (else
+          ;; (display "#<unknown>" out)
+          (write x out))))
+      (check-shared x "" wr-one))))
 
 (define write/ss write-with-shared-structure)
 
@@ -171,6 +187,9 @@
            (read-char in)
            (read in)
            (skip-whitespace-and-sexp-comments in read))
+          ((eqv? #\| (peek-char in))
+           (skip-comment in 0)
+           (skip-whitespace-and-sexp-comments in read))
           (else #f)))
    (else
     #t)))
@@ -189,6 +208,34 @@
     ("escape" . ,(integer->char 27))
     ("delete" . ,(integer->char 127))))
 
+(define U1 1)
+(define S8 2)
+(define U8 3)
+(define S16 4)
+(define U16 5)
+(define S32 6)
+(define U32 7)
+(define S64 8)
+(define U64 9)
+(define F32 10)
+(define F64 11)
+(define C64 12)
+(define C128 13)
+(define F8 14)
+(define F16 15)
+
+(define (resolve-uniform-type c prec)
+  (or
+   (case prec
+     ((1) (and (eqv? c #\u) U1))
+     ((8) (case c ((#\u) U8) ((#\s) S8) ((#\f) F8) (else #f)))
+     ((16) (case c ((#\u) U16) ((#\s) S16) ((#\f) F16) (else #f)))
+     ((32) (case c ((#\u) U32) ((#\s) S32) ((#\f) F32) (else #f)))
+     ((64) (case c ((#\u) U64) ((#\s) S64) ((#\f) F64) ((#\c) C64) (else #f)))
+     ((128) (case c ((#\c) C128) (else #f)))
+     (else #f))
+   (error "invalid uniform type" c prec)))
+
 (define read-with-shared-structure
   (let ((read read))
     (lambda o
@@ -370,13 +417,30 @@
             ((#\() (list->vector (read-one in)))
             ((#\') (read-char in) (list 'syntax (read-one in)))
             ((#\`) (read-char in) (list 'quasisyntax (read-one in)))
-            ((#\t) (let ((s (read-name #f in)))
-                     (or (string-ci=? s "t") (string-ci=? s "true")
-                         (read-error "bad # syntax" s))))
-            ((#\f) (let ((s (read-name #f in)))
-                     (if (or (string-ci=? s "f") (string-ci=? s "false"))
-                         #f
-                         (read-error "bad # syntax" s))))
+            ((#\,) (read-char in)
+             (let ((sym (if (eqv? #\@ (peek-char in))
+                            (begin (read-char in) 'unsyntax-splicing)
+                            'unsyntax)))
+               (list sym (read-one in))))
+            ((#\t)
+             (let ((s (read-name #f in)))
+               (or (string-ci=? s "t") (string-ci=? s "true")
+                   (read-error "bad # syntax" s))))
+            ((#\f)
+             (let ((s (read-name #f in)))
+               (cond
+                ((or (string-ci=? s "f") (string-ci=? s "false"))
+                 #f)
+                ((member s '("f8" "F8"))
+                 (list->uvector F8 (read in)))
+                ((member s '("f16" "F16"))
+                 (list->uvector F16 (read in)))
+                ((member s '("f32" "F32"))
+                 (list->uvector F32 (read in)))
+                ((member s '("f64" "F64"))
+                 (list->uvector F64 (read in)))
+                (else
+                 (read-error "bad # syntax" s)))))
             ((#\d) (read-char in) (read in))
             ((#\x) (read-char in) (read-number 16))
             ((#\o) (read-char in) (read-number 8))
@@ -386,21 +450,16 @@
              (let ((s (read-name #\# in)))
                (or (string->number s)
                    (read-one (open-input-string (substring s 2))))))
-            ((#\u #\v)
-             (if (eqv? #\v (peek-char in))
+            ((#\u #\v #\s #\c)
+             (if (char-ci=? #\v (peek-char in))
                  (read-char in))
-             (read-char in)
-             (if (not (eqv? #\8 (peek-char in)))
-                 (read-error "invalid syntax #u" (peek-char in)))
-             (read-char in)
-             (let ((ls (read-one in)))
+             (let* ((c (char-downcase (read-char in)))
+                    (prec (read-number 10))
+                    (etype (resolve-uniform-type c prec))
+                    (ls (read-one in)))
                (if (not (list? ls))
-                   (read-error "invalid bytevector syntax" ls))
-               (let* ((len (length ls))
-                      (bv (make-bytevector len)))
-                 (do ((i 0 (+ i 1)) (ls ls (cdr ls)))
-                     ((null? ls) bv)
-                   (bytevector-u8-set! bv i (car ls))))))
+                   (read-error "invalid uniform vector syntax" ls))
+               (list->uvector etype ls)))
             ((#\\)
              (read-char in)
              (let* ((c1 (read-char in))
